/*
 * main.c
 * Liberté
 *
 * Copyright (c) 2005 Landon J. Fuller <landonf@opendarwin.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <Foundation/Foundation.h>

#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <errno.h>
#include <sys/sysctl.h>

#include <stdio.h>

#include "Liberte.h"
#include "Shout.h"

#import "LEPreferences.h"

#define MDNSRESPONDER_NAME "mDNSResponder"
#define MDNSRESPONDER_UID 0

/* Find a pid by name and uid */
static pid_t find_process(const char *name, uid_t uid) {
	int mib[5];
	struct kinfo_proc *procs = NULL, *newprocs;
	int i, st, nprocs;
	size_t miblen, size;

	/* Set up sysctl MIB */
	mib[0] = CTL_KERN;
	mib[1] = KERN_PROC;
	mib[2] = KERN_PROC_UID;
	mib[3] = uid;
	mib[3] = 0;     
	miblen = 4;

	/* Get initial sizing */
	st = sysctl(mib, miblen, NULL, &size, NULL, 0);

	/* Repeat until we get them all ... */
	do {
		/* Room to grow */
		size += size / 10;
		newprocs = realloc(procs, size);
		if (!newprocs) {
			if (procs) {
				free(procs);
			}
			NSLog(@"realloc failed.");
			return (0);
		}
		procs = newprocs;
		st = sysctl(mib, miblen, procs, &size, NULL, 0);
	} while (st == -1 && errno == ENOMEM);

	if (st != 0) {
		NSLog(@"sysctl(KERN_PROC) failed.");
		return (0);
	}

	/* Do we match the kernel? */
	assert(size % sizeof(struct kinfo_proc) == 0);

	nprocs = size / sizeof(struct kinfo_proc);

	if (!nprocs) {
		NSLog(@"No processes found.");
		return(0);
	}

	for (i = 0; i < nprocs; i++) {
		if (strncmp(procs[i].kp_proc.p_comm, name, MAXCOMLEN) == 0) {
			pid_t pid = procs[i].kp_proc.p_pid;
			free(procs);
			return (pid);
		}
	}
	free(procs);
	return (0);
}

/* Load our Shout payload */
static void *load_shout (void) {
	NSString *appSupportSubpath = LIBERTE_APP_SUPPORT_DIR;
	NSArray *librarySearchPaths;
	NSEnumerator *searchPathEnum;
	NSString *currPath;

	/* Locate our Application Support directory */
	librarySearchPaths = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSLocalDomainMask, YES);
	searchPathEnum = [librarySearchPaths objectEnumerator];
	while(currPath = [searchPathEnum nextObject]) {
		currPath = [currPath stringByAppendingPathComponent: appSupportSubpath];
		currPath = [currPath stringByAppendingPathComponent: LIBERTE_BUNDLE_NAME];
		if ([[NSFileManager defaultManager] fileExistsAtPath: currPath isDirectory: NULL]) {
			CFBundleRef injectionBundle = NULL;
			CFURLRef injectionURL = NULL;
			void *injectionCode = NULL;
			
			injectionURL = (CFURLRef) [NSURL fileURLWithPath: currPath];
			if (!(injectionBundle = CFBundleCreate(kCFAllocatorDefault, injectionURL))) {
				NSLog(@"Could not create an injection bundle instance.");
				return (NULL);
			}
		
			/* Code Injection */
			if (!(injectionCode = CFBundleGetFunctionPointerForName(injectionBundle, CFSTR(INJECT_ENTRY_SYMBOL)))) {
				NSLog(@"Could not find injection entry point in bundle.");
				CFRelease(injectionBundle);
				return (NULL);
			}
			return (injectionCode);
		}
	}
	
	NSLog(@"%@ was not found.", LIBERTE_BUNDLE_NAME);
	return (NULL);
}

int main (int argc, char *argv[]) {
	NSAutoreleasePool	*pool = [[NSAutoreleasePool alloc] init];
	const char			*multicastGroup;
	LEPreferences		*prefs;
	void				*injectionCode;
	ShoutConfig			config;
	mach_error_t		err;
	int					n;
	pid_t				pid;
	
	/* Initialize Preferences */
	prefs =  [[LEPreferences alloc] init];

	/* Load Injection Bundle */
	if (!(injectionCode = load_shout())) {
		NSLog(@"Failed to load injection bundle.");
		exit(1);
	}
	
	/* Find mDNSResponder Process */
	if (!(pid = find_process(MDNSRESPONDER_NAME, MDNSRESPONDER_UID))) {
		NSLog(@"Failed to find mDNSResponder process.");
		exit(1);
	}

	/* Pull multicast group from preferences */
	if (!(multicastGroup = [[prefs multicastGroup] UTF8String])) {
		NSLog(@"Failed to load preferences");
		exit(1);
	}
	
	/* Try IPv4 first ... */
	if ((n = inet_pton(AF_INET, multicastGroup, &config.multicastGroup.v4)) == 0) {
		NSLog(@"Error in inet_pton: %s", strerror(errno));
		exit(1);
	} else if (n == 1) {
		/* Success */
		config.multicastGroup.family = AF_INET;
	}

	/* If IPv4 failed, try IPv6 */
	if (n == 0) {
		if((n = inet_pton(AF_INET6, multicastGroup, &config.multicastGroup.v6)) < 0) {
			NSLog(@"Error in inet_pton: %s", strerror(errno));
			exit(1);
		} else if (n == 0) {
			NSLog(@"Invalid multicast group specified: %s", multicastGroup);
			exit(1);
		} else {
			/* Success */
			config.multicastGroup.family = AF_INET6;
			// TODO: IPv6
			NSLog(@"IPv6 Multicast is not supported");
			exit(1);
		}
	}

	if((err = mach_inject(injectionCode, (const void *) &config, sizeof(config), pid, 0)) != 0) {
		NSLog(@"mach_inject() failure. pid=%d, err=%d", pid, err);
		exit(1);
	}

	[prefs release];
	[pool release];
	exit (0);
}
