/*
 * mDNSHooks.m
 * Liberté
 *
 * Copyright (c) 2005 Landon J. Fuller <landonf@opendarwin.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <syslog.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/param.h> /* For MAX() macro */
#include <netinet/in.h>
#include <mach-o/dyld.h>

#include "safe_malloc.h"

#include "mDNSHooks.h"
#include "mDNSEmbeddedAPI.h"

/* 0xE00000FB == 224.0.0.251, the default mDNS multicast address */
#define MDNS_MCAST_GROUP 0xE00000FB

/*
 * Global Initialization
 */

static ShoutConfig *config;

/* Unmodified mDNS Functions */
#define mDNS_LockSymbol "_mDNS_Lock"
typedef void (*mDNS_LockPtr)(mDNS *const m);
static mDNS_LockPtr mDNS_Lock;

#define mDNS_UnlockSymbol "_mDNS_Unlock"
typedef void (*mDNS_UnlockPtr)(mDNS *const m);
static mDNS_UnlockPtr mDNS_Unlock;

#define mDNSMacOSXNetworkChangedSymbol "_mDNSMacOSXNetworkChanged"
typedef void (*mDNSMacOSXNetworkChangedPtr)(mDNS *const m);
static mDNSMacOSXNetworkChangedPtr mDNSMacOSXNetworkChanged;

void mDNSHooksInitialize(ShoutConfig *configPtr) {
	config = configPtr;
	_dyld_lookup_and_bind(mDNS_LockSymbol, (void *) &mDNS_Lock, NULL);
	_dyld_lookup_and_bind(mDNS_UnlockSymbol, (void *) &mDNS_Unlock, NULL);
	_dyld_lookup_and_bind(mDNSMacOSXNetworkChangedSymbol, (void *) &mDNSMacOSXNetworkChanged, NULL);
}

/*
 * mDNS_Execute()
 * Insert ourselves into the mDNSResponder event loop.
 * On first run, we call mDNSMacOSXNetworkChanged(), which forces all sockets to be closed and re-opened. This allows
 * us to join sockets to our multicast group.
 */

static mDNS_ExecutePtr mDNS_ExecuteOrig;

void mDNS_ExecuteInitialize(mDNS_ExecutePtr func) {
	mDNS_ExecuteOrig = func;
}

#define MDNS_SECURE_USERNAME "nobody"
#define MDNS_SECURE_GROUP "nogroup"
#define MDNS_SECURE_CHROOT "/var/empty"

/* Secure the mDNS daemon */
static int mDNSSecureDaemon(void) {
	struct passwd user;
	struct passwd *pwbufp;
	struct group group;
	struct group *grbufp;
	size_t uLen, gLen, len;
	int error;
	char *buf;
	
	/* Get proper buffer sizes */ 
	uLen = sysconf(_SC_GETPW_R_SIZE_MAX);
	gLen = sysconf(_SC_GETGR_R_SIZE_MAX);
#if MACOSX_NON_SUCK
	len = MAX(uLen, gLen);
#endif
	// XXX - We have to fudge it;
	len = 1024;

	/* ... and allocate our string buffer */
	buf = safe_malloc(len);
	
	/* Look up our user */
	if ((error = getpwnam_r(MDNS_SECURE_USERNAME, &user, buf, len, &pwbufp)) != 0) {
		printf("Internal Liberte Error in %s, %d: %s\n", __FUNCTION__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	/* ... and the group */
	if ((error = getgrnam_r(MDNS_SECURE_GROUP, &group, buf, len, &grbufp)) != 0) {
		printf("Internal Liberte Error in %s, %d: %s\n", __FUNCTION__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	/* Start with a chroot() */
	/* I believe chroot() does this nowdays, but just to be safe ...*/
	if (chdir(MDNS_SECURE_CHROOT) != 0) {
		printf("Internal Liberte Error in %s, %s:%d: %s\n", __FUNCTION__, __FILE__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	if (chroot(MDNS_SECURE_CHROOT) != 0) {
		printf("Internal Liberte Error in %s, %s:%d: %s\n", __FUNCTION__, __FILE__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	/* Drop our supplementary groups */
	if (setgroups(0, NULL) != 0) {
		printf("Internal Liberte Error in %s, %s:%d: %s\n", __FUNCTION__, __FILE__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	/* Drop our root gid */
	if (setgid(group.gr_gid) != 0) {
		printf("Internal Liberte Error in %s, %s:%d: %s\n", __FUNCTION__, __FILE__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	/* Finally, drop our uid */
	if (setuid(user.pw_uid) != 0) {
		printf("Internal Liberte Error in %s, %s:%d: %s\n", __FUNCTION__, __FILE__, __LINE__, strerror(errno));
		free(buf);
		return (0);
	}
	
	/* Hey, it worked! */
	free(buf);
	return (1);
}

mDNSs32 mDNS_ExecuteOverride(mDNS *const m) {
	static int initDone = 0;
	
	if (!initDone) {
		mDNSMacOSXNetworkChanged(m);
		mDNSSecureDaemon();
		initDone = 1;
	}
	
	return (mDNS_ExecuteOrig(m));
}

/*
 * mDNSAddrIsDNSMulticast()
 * Modify the function to return success when passed our multicast address.
 */
static mDNSAddrIsDNSMulticastPtr mDNSAddrIsDNSMulticastOrig;

void mDNSAddrIsDNSMulticastInitialize(mDNSAddrIsDNSMulticastPtr func) {
	mDNSAddrIsDNSMulticastOrig = func;
}

mDNSexport mStatus mDNSAddrIsDNSMulticastOverride(const mDNSAddr *ip) {
	// TODO: IPv6
	switch (ip->type) {
		case mDNSAddrType_IPv4:
			if (ip->ip.v4.NotAnInteger == config->multicastGroup.v4.s_addr) {
				return (mDNStrue);
			} else {
				return (mDNSAddrIsDNSMulticastOrig(ip));
			}
		default:
			return (mDNSAddrIsDNSMulticastOrig(ip));
	}
}


/*
 * mDNSPlatformSendUDP()
 * If the destination is the standard mDNS multicast address,
 * also send the packet to our multicast address.
 */
static mDNSPlatformSendUDPPtr mDNSPlatformSendUDPOrig;

void mDNSPlatformSendUDPInitialize(mDNSPlatformSendUDPPtr func) {
	mDNSPlatformSendUDPOrig = func;
}

mDNSexport mStatus mDNSPlatformSendUDPOverride(const mDNS *const m,
					       const void *const msg,
					       const mDNSu8 *const end,
					       mDNSInterfaceID InterfaceID,
					       const mDNSAddr *dst,
					       mDNSIPPort dstPort)
{
	mDNSAddr siteDest;

	// TODO: IPv6
	if (dst->type != mDNSAddrType_IPv4) {
		return (mDNSPlatformSendUDPOrig(m, msg, end, InterfaceID, dst, dstPort));
	}

	/* Is the packet destined for the standard multicast group? */
	if (dst->ip.v4.NotAnInteger == MDNS_MCAST_GROUP) {
		/* Transmit to our user-specified multicast group */
		siteDest.type = mDNSAddrType_IPv4;
		siteDest.ip.v4.NotAnInteger = config->multicastGroup.v4.s_addr;
		mDNSPlatformSendUDPOrig(m, msg, end, InterfaceID, &siteDest, dstPort);
	}

	/* Standard Dispatch */	
	return (mDNSPlatformSendUDPOrig(m, msg, end, InterfaceID, dst, dstPort));
}